
 {**********************************************************************}
 {                                                                      }
 {   Easy PCX  -  A freeware library to show/make PCX files             }
 {                using the BGI.                                        }
 {                It works in ANY graphic mode having 2, 16             }
 {                or 256 colors (SVGA too!).                            }
 {                                                                      }
 {   Author:  Pino Navato                                               }
 {            Via Pittore, 164                                          }
 {            80046 S.Giorgio a Cremano (NA)                            }
 {            ITALY                                                     }
 {                                                                      }
 {   Acknowledgment:  This code includes small portions of SHOW_PCX.    }
 {                    Used by permission of ZSoft Corporation.          }
 {                                                                      }
 {**********************************************************************}


Unit EasyPCX;
{$I-}

interface

   const English  = 0;    { Da usare con PCXerrorMsg }
         Italiano = 1;

   const active    = 1;   { Da usare con ShowPCX }
         disactive = 0;

   const PCXok                 = 0;   { Codici d'errore }
         PCXcannotOpen         = 1;
         PCXreadError          = 2;
         PCXinvalidFile        = 3;
         PCXunknownVersion     = 4;
         PCXunknownCompression = 5;
         PCXpaletteNotFound    = 6;
         PCXunknownMemoryModel = 7;
         PCXnotEnoughColors    = 8;
         PCXoutOfMemory        = 9;
         PCXwriteError         = 10;

   const REDc   = 0;
         GREENc = 1;
         BLUEc  = 2;

   type  RGB256Type       = array[0..255, REDc..BLUEc] of byte;
         EGA16paletteType = array[0..15] of byte;
         PCXinfoRec       = record
                               Hres, Vres   : word;
                               BitsPerPixel : byte;
                               x1, y1       : word;
                               version      : byte;
                               RGBpalette   : RGB256Type;
                               EGA16palette : EGA16paletteType
                            end;


   function  PCXresult: byte;
   function  PCXerrorMsg(ErrorCode, language: byte): string;
   procedure AssignPCX(filename: string);
   procedure GetPCXinfo(var PCXinfo: PCXinfoRec);
   procedure ClosePCX;
   procedure ShowPCX(AutoCentering, PaletteUsage: byte; BitBlt: byte);
   procedure MakePCX(X1, Y1, X2, Y2: word);



implementation
   uses DOS, Graph;


   { *ENG*  The following banner will never appear on the screen, it will
            simply stay inside EXE files.  Please, PLEASE, do not change it! }
   { *ITA*  Il banner seguente non apparir mai sullo schermo, bens soltanto
            all'interno dei file EXE.  Per favore, PER FAVORE, non cambiarlo! }
   const banner: array[1..86] of char =
            #13+#10 + #13+#10 +
            '      This program uses   EASY PCX   a freeware library by Pino Navato      ' +
            #13+#10 + #13+#10 ;


   type PCXheader = record
                       Manufacturer : byte;   { Always 10 for PCX files }

                       Version      : byte;   { 2 - with palette,
                                                3 - no palette,
                                                4 - Old MS-Windows files - no palette,
                                                5 - with palette }

                       Encoding     : byte;   { 1 is PCX, it is possible that we add
                                                additional encoding methods in the future }

                       Bits_per_pixel_per_plane : byte;   { Number of bits to represent a pixel
                                                            (per plane) - 1, 2, 4, or 8 }

                       Xmin    : word;        { Image window dimensions (inclusive) }
                       Ymin    : word;        { Xmin, Ymin are usually zero (not always) }
                       Xmax    : word;
                       Ymax    : word;

                       Hdpi    : word;        { Resolution of image (dots per inch) }
                       Vdpi    : word;        { Set to scanner resolution - 300 is default }

                       ColorMap16: array [0..15, REDc..BLUEc] of byte;
                                              { RGB palette data (16 colors or less)
                                                256 color palette is appended to end of file }

                       Reserved: byte;        { (used to contain video mode)
                                                now it is ignored - just set to zero }

                       Nplanes : byte;        { Number of planes }

                       Bytes_per_line_per_plane: word;   { Number of bytes to allocate
                                                           for a scanline plane.
                                                           MUST be an an EVEN number!
                                                           Do NOT calculate from Xmax-Xmin! }

                       PaletteInfo: word;     { 1 = black & white or color image,
                                                2 = grayscale image - ignored in PB4, PB4+
                                                palette must also be set to shades of gray! }

                       HscreenSize: word;     { added for PC Paintbrush IV Plus ver 1.0,  }
                       VscreenSize: word;     { PC Paintbrush IV ver 1.02 (and later)     }
                                              { These fields will probably just contain
                                                garbage. }

                       Filler: array [74..127] of byte;     { Just set to zeros }
                    end;


   const PCXnotAssigned   = 250;    { Run-Time Errors codes }
         NotInGraphMode   = 251;
         InvalidGraphMode = 252;
         InvalidCoords    = 253;
         InvalidParameter = 254;


   var PCXfile     : file;
       PCXopen,
       PCXassigned : boolean;
       PCXerrCode  : byte;

   var header      : PCXheader;
       ColorMap256 : RGB256Type;
       EGA16palette: EGA16paletteType;    { Impostata da EvalEGApalette }

   var bits_per_pixel : byte;
       bytes_per_line : word;
       Hres, Vres     : word;



   function ImageSize(x1, y1, x2, y2: Integer): Word;  { Calcola la misura ESATTA }
      { * Vedi nota 1! * }
      var w,h  : word;
          size : word;  { Dim. in byte di un piano di bit }
      begin
         w:=x2-x1+1;
         h:=y2-y1+1;
         case GetMaxColor of
              1: begin
                   size:= w shr 3;                   { size:= w div 8 }
                   if w and 7 <> 0 then inc(size);   { if size mod 8 <> 0 then... }
                   ImageSize:=size * h + 4
                 end;
             15: begin
                   size:= w shr 3;                   { size:= w div 8 }
                   if w and 7 <> 0 then inc(size);   { if size mod 8 <> 0 then... }
                   ImageSize:=size*4 * h + 4;  { 4 piani di size bytes ciascuno * h + l'header }
                 end;
            255: ImageSize:= w*h+4;
         else
            RunError(255)
         end { case }
      end;



   function PCXresult: byte;
     begin
        PCXresult:=PCXerrCode
     end;


   function PCXerrorMsg(ErrorCode, language: byte): string;
      const Msg: array[0..10, 0..1] of string[34] =
               (('No error',                    'Nessun errore'),
                ('Cannot open file',            'Impossibile aprire il file'),
                ('Read error',                  'Errore di lettura'),
                ('Not a PCX file',              'File non PCX'),
                ('Unknown PCX version',         'Versione di PCX sconosciuta'),
                ('Unknown compression method',  'Metodo di compressione sconosciuto'),
                ('Palette not found',           'Palette non trovata'),
                ('Unknown memory model',        'Modello di memoria sconosciuto'),
                ('Not enough colors',           'Colori insufficienti'),
                ('Out of memory',               'Memoria insufficiente'),
                ('Write error',                 'Errore di scrittura'));

      begin
         if (ErrorCode>10) or (language>1) then
            RunError(InvalidParameter)
         else
            PCXerrorMsg:=Msg[ErrorCode, language]
      end;


   procedure AssignPCX(filename: string);
     begin
        PCXerrCode:=0;
        if PCXopen then ClosePCX;
        assign(PCXfile, filename);
        PCXassigned:=true
     end;


   procedure ClosePCX;
     begin
        PCXerrCode:=0;
        if PCXopen then
           begin
              close(PCXfile);
              if IOresult<>0 then PCXerrCode:=PCXwriteError;   { V. nota 2 }
              PCXopen:=false
           end
     end;


   procedure ReadHeader;    { Apre il file e legge l'header }
     begin
        PCXerrCode:=0;
        if not PCXassigned then RunError(PCXnotAssigned);     { Errore fatale! }
        if PCXopen then ClosePCX;                  { Se era aperto lo chiude, poi lo riaprir }
        reset(PCXfile, 1);
        if IOresult<>0 then
           PCXerrCode:=PCXcannotOpen               { Errore non fatale }
        else
           begin
              PCXopen:=true;                       { File aperto con successo }
              BlockRead(PCXfile, header, 128);
              if IOresult<>0 then
                 begin
                    ClosePCX;
                    PCXerrCode:=PCXreadError
                 end
           end
     end;


   procedure CheckHeader;
     begin
        if header.Manufacturer<>10 then PCXerrCode:=PCXinvalidFile
        else if not (header.Version in [0..5]) then PCXerrCode:=PCXunknownVersion
        else if header.Encoding<>1 then PCXerrCode:=PCXunknownCompression
     end;


   procedure EvalEGApalette;    { V. nota 3 }
     var i, r, g, b: integer;
     begin
        for i := 0 to 15 do
           begin
              r := Header.ColorMap16 [i, REDc]   shr 1;
              g := Header.ColorMap16 [i, GREENc] shr 2;
              b := Header.ColorMap16 [i, BLUEc]  shr 3;
              EGA16palette [i] := (r and 32) or (g and 16) or (b and 8);
              r := Header.ColorMap16 [i, REDc]   shr 5;
              g := Header.ColorMap16 [i, GREENc] shr 6;
              b := Header.ColorMap16 [i, BLUEc]  shr 7;
              EGA16palette [i] := EGA16palette [i] or (r and 4) or (g and 2) or b
           end
     end;


   procedure ReadColorMap256;
     var b: byte;
     begin
        if FileSize(PCXfile) < 128+769 then
           begin
              ClosePCX;
              PCXerrCode:=PCXpaletteNotFound;
              exit
           end;
        seek(PCXfile, fileSize(PCXfile)-769);
        BlockRead(PCXfile, b, 1);          { Read palette indicator }
        if IOresult<>0 then
          begin
              ClosePCX;
              PCXerrCode:=PCXreadError;
              exit
           end;
        if b<>12 then PCXerrCode:=PCXpaletteNotFound     { No palette here! }
        else
           begin
              BlockRead(PCXfile, ColorMap256, 256*3);
              if IOresult<>0 then
                 begin
                    ClosePCX;
                    PCXerrCode:=PCXreadError;
                    exit
                 end;
              seek(PCXfile, 128);          { Go back to start of PCX data }
              if IOresult<>0 then
                 begin
                    ClosePCX;
                    PCXerrCode:=PCXreadError;
                    exit
                 end
           end
     end;


   procedure GetPCXinfo(var PCXinfo: PCXinfoRec);
     begin
        PCXerrCode:=0;
        FillChar(PCXinfo, SizeOf(PCXinfo), 0);   { Inizializza PCXinfo }
        ReadHeader;
        if PCXresult<>PCXok then exit;           { Errore di lettura }
        CheckHeader;
        if PCXresult=PCXinvalidFile then exit;   { Non  un file PCX: inutile riempire PCXinfo }
        with PCXinfo do
           begin
              Hres:=header.Xmax-header.Xmin+1;
              Vres:=header.Ymax-header.Ymin+1;
              BitsPerPixel:=header.bits_per_pixel_per_plane*header.Nplanes;
              x1:=header.Xmin;
              y1:=header.Ymin;
              version:=header.version
           end;
        if PCXresult<>PCXok then exit;  {CheckHeader aveva trovato un errore: inutile procedere}
        if header.version in [2,5] then       { C' la palette }
           begin
              if PCXinfo.BitsPerPixel=4 then     { 16 colori }
                 begin
                    move(header.ColorMap16, PCXinfo.RGBpalette, 16*3);
                    EvalEGApalette;
                    move(EGA16palette, PCXinfo.EGA16palette, 16)
                 end;
              if PCXinfo.BitsPerPixel=8 then     { 256 colori }
                 begin
                    ReadColorMap256;
                    if PCXresult=PCXok then
                       move(ColorMap256, PCXinfo.RGBpalette, 256*3)
                 end
           end
     end;



  procedure SetPalette(colors: word);
    var Reg    : registers;
        gd, gm : integer;

    procedure SetEGApalette;
      var i      : byte;
          palette: PaletteType;
      begin
         with palette do
            begin
               size:=16;
               for i := 0 to 15 do
                   colors[i]:=EGA16palette[i]
            end;
         SetAllPalette(palette)
      end;

    procedure SetPalette16;
      var i      : byte;
          palette: PaletteType;
      begin
         with palette do
            begin
               size:=16;
               for i := 0 to 15 do
                   colors[i]:=i
            end;
         SetAllPalette(palette);

         for i := 0 to 15 do
            begin                                { R, G, and B must be 0..63 }
               ColorMap256 [i, REDc]   := Header.ColorMap16 [i, REDc]   shr 2;
               ColorMap256 [i, GREENc] := Header.ColorMap16 [i, GREENc] shr 2;
               ColorMap256 [i, BLUEc]  := Header.ColorMap16 [i, BLUEc]  shr 2
            end;

         Reg.ah := $10;                      { Set DAC Call }
         Reg.al := $12;                      { set a block of DAC registers }
         Reg.bx := 0;                        { first DAC register number }
         Reg.cx := 16;                       { number of registers to update }
         Reg.dx := ofs (ColorMap256);        { offset of block }
         Reg.es := seg (ColorMap256);        { segment of block }
         intr ($10, Reg)                     { call interrupt }
      end;

    procedure SetPalette256(UsedColors: word);   { UsedColors = colori realm. usati (16 o 256) }
      var i: byte;
      begin
         for i := 0 to UsedColors-1 do
            begin                                { R, G, and B must be 0..63 }
               ColorMap256 [i, REDc]   := ColorMap256 [i, REDc]   shr 2;
               ColorMap256 [i, GREENc] := ColorMap256 [i, GREENc] shr 2;
               ColorMap256 [i, BLUEc]  := ColorMap256 [i, BLUEc]  shr 2
            end;

         Reg.ah := $10;                      { Set DAC Call }
         Reg.al := $12;                      { set a block of DAC registers }
         Reg.bx := 256-UsedColors;           { first DAC register number }
         Reg.cx := UsedColors;               { number of registers to update }
         Reg.dx := ofs (ColorMap256);        { offset of block }
         Reg.es := seg (ColorMap256);        { segment of block }
         intr ($10, Reg)                     { call interrupt }
      end;


    begin { SetPalette }
       if GetMaxColor=255 then
          begin
             if colors=16 then
                move(header.ColorMap16, ColorMap256, 16*3);
             SetPalette256(colors)
          end
       else   { GetMaxColor=15 }
          begin
             DetectGraph(gd,gm);
             if (gd>=3) and (gd<=5) then  SetEGApalette
             else SetPalette16
          end
    end;



   const SecurityBlock = 63+4;   { Bytes da allocare sull'heap in pi rispetto al min. indisp. }
         BlockBufLen   = 4096;   { Bytes da leggere con BlockRead }

   type  PCXlineBufferType = array [-4..32767] of byte;
         ShowLineProc      = procedure (X, Y: Integer; var BitMap; BitBlt: Word);

   var BlockBuffer   : array[0..BlockBufLen-1] of byte;  { Conterr i byte letti con BlockRead }
       NextByte      : word;                             { Indice all'interno di BlockBuffer[] }
       NumBlocksRead : word;                             { Bytes letti con BlockRead }
       PCXlineBufLen : word;
       PCXlineBuffer : ^PCXlineBufferType;  { Conterr la bitmap decompressa }
       index         : word;                { Indice all'interno di PCXlineBuffer^[] }
       data          : byte;                { Usata per leggere da PCXfile }
       ImageBufLen   : word;
       ImageBuffer   : ^PCXlineBufferType;  { Usato per la conversione "con travaso" }
       PlaneLength   : word;    { BytesPerLinePerPlane per immagini a 2 o 16 col.
                                  che superano il margine dx dello schermo }



    procedure DecompressLine;
      var count: byte;

      function ReadByte: byte;  { Read a byte from PCXfile.  Use BlockRead because it is FAST! }
        begin
           if NextByte = NumBlocksRead then      { Il buffer si  esaurito }
              begin
                 BlockRead (PCXfile, BlockBuffer, BlockBufLen, NumBlocksRead);
                 if (IOresult<>0) or (NumBlocksRead=0) then
                    begin
                       PCXerrCode:=PCXreadError;
                       NumBlocksRead:=BlockBufLen   { Per evitare altre letture }
                    end;
                 NextByte := 0
              end;
           ReadByte := BlockBuffer[NextByte];    { Preleva il prossimo byte dal buffer }
           inc (NextByte)                        { NextByte++; }
        end;  { ReadByte }

      begin { DecompressLine }
         { First of all bring in any data that wrapped from previous line }
         { usually none  -  this is just to be safe                       }
         if Index <> 0 then
            FillChar (PCXlineBuffer^[0], Index, data);    { fills a contiguous block of data }

         while (Index < bytes_per_line) do          { read 1 line of data (all planes) }
            begin
               data := ReadByte;
               if (data and $C0) = $C0 then    { Questo byte contiene il numero di ripetizioni }
                  begin
                     count := data and $3F;    { Preleva il numero di ripetizioni }
                     data := ReadByte;         { Legge il byte da ripetere "count" volte }
                     FillChar (PCXlineBuffer^[Index], count, data);  { fills a block }
                     inc (Index, count)                              { Index += count; }
                  end
               else
                  begin
                     PCXlineBuffer^[Index] := data;
                     inc (Index)                                     { Index++; }
                  end
            end; { while }
         dec(Index,bytes_per_line)  {Index  >0 se sono avanzati dei byte per la prossima riga}
      end;



  procedure Show16_16(X, Y: Integer; var BitMap; BitBlt: Word); FAR;  { V. nota 4 }
    var base: word;
    begin
       base:=header.bytes_per_line_per_plane*3+PlaneLength;
       move(PCXlineBuffer^[header.bytes_per_line_per_plane*2],
            PCXlineBuffer^[base], PlaneLength);
       move(PCXlineBuffer^[header.bytes_per_line_per_plane],
            PCXlineBuffer^[base+PlaneLength], PlaneLength);
       move(PCXlineBuffer^[0],
            PCXlineBuffer^[base+PlaneLength*2], PlaneLength);
       move(PCXlineBuffer^[-4], PCXlineBuffer^[header.bytes_per_line_per_plane*3-4], 4);
       PutImage(x,y, PCXlineBuffer^[header.bytes_per_line_per_plane*3-4], BitBlt)
    end;


  procedure Show16_256(X, Y: Integer; var BitMap; BitBlt: Word); FAR;
    var i,k: word;
        b1,b2,b3,b4: byte;
    begin
       for i:=0 to PlaneLength-1 do
          begin
             b4:=PCXlineBuffer^[i];
             b3:=PCXlineBuffer^[header.bytes_per_line_per_plane+i];
             b2:=PCXlineBuffer^[header.bytes_per_line_per_plane*2+i];
             b1:=PCXlineBuffer^[header.bytes_per_line_per_plane*3+i];
             k:=i shl 3;   { k:=i*8 }
             ImageBuffer^[k]:=(b1 and 128 shr 4) or (b2 and 128 shr 5) or (b3 and 128 shr 6) or
                                                                              (b4 shr 7) or 240;
             ImageBuffer^[k+1]:=(b1 and 64 shr 3) or (b2 and 64 shr 4) or (b3 and 64 shr 5) or
                                                                       (b4 and 64 shr 6) or 240;
             ImageBuffer^[k+2]:=(b1 and 32 shr 2) or (b2 and 32 shr 3) or (b3 and 32 shr 4) or
                                                                       (b4 and 32 shr 5) or 240;
             ImageBuffer^[k+3]:=(b1 and 16 shr 1) or (b2 and 16 shr 2) or (b3 and 16 shr 3) or
                                                                       (b4 and 16 shr 4) or 240;
             ImageBuffer^[k+4]:=(b1 and 8) or (b2 and 8 shr 1) or (b3 and 8 shr 2) or
                                                                        (b4 and 8 shr 3) or 240;
             ImageBuffer^[k+5]:=(b1 and 4 shl 1) or (b2 and 4) or (b3 and 4 shr 1) or
                                                                        (b4 and 4 shr 2) or 240;
             ImageBuffer^[k+6]:=(b1 and 2 shl 2) or (b2 and 2 shl 1) or (b3 and 2) or
                                                                        (b4 and 2 shr 1) or 240;
             ImageBuffer^[k+7]:=(b1 and 1 shl 3) or (b2 and 1 shl 2) or (b3 and 1 shl 1) or
                                                                               (b4 and 1) or 240
          end;
       PutImage(x,y, ImageBuffer^, BitBlt)
    end;


  procedure Show2_16(X, Y: Integer; var BitMap; BitBlt: Word); FAR;
    begin
       move(PCXlineBuffer^[0], PCXlineBuffer^[PlaneLength], PlaneLength);
       move(PCXlineBuffer^[0], PCXlineBuffer^[PlaneLength*2], PlaneLength*2);
       PutImage(x,y, PCXlineBuffer^, BitBlt)
    end;


  procedure Show2_256(X, Y: Integer; var BitMap; BitBlt: Word); FAR;
    var i,k:word;
    begin
       for i:=0 to PlaneLength-1 do
          begin
             k:=i shl 3;   { k:=i*8 }
             case PCXlineBuffer^[i] of
                0: FillChar(ImageBuffer^[k],8,0);    { Converte 8 pixel neri in un colpo solo }
              255: FillChar(ImageBuffer^[k],8,15)    { Converte 8 pixel bianchi ...           }
              else
                 begin
                    if PCXlineBuffer^[i] and 128=128 then
                       ImageBuffer^[k]:=15
                    else
                       ImageBuffer^[k]:=0;
                    if PCXlineBuffer^[i] and 64=64 then
                       ImageBuffer^[k+1]:=15
                    else
                       ImageBuffer^[k+1]:=0;
                    if PCXlineBuffer^[i] and 32=32 then
                       ImageBuffer^[k+2]:=15
                    else
                       ImageBuffer^[k+2]:=0;
                    if PCXlineBuffer^[i] and 16=16 then
                       ImageBuffer^[k+3]:=15
                    else
                       ImageBuffer^[k+3]:=0;
                    if PCXlineBuffer^[i] and 8=8 then
                       ImageBuffer^[k+4]:=15
                    else
                       ImageBuffer^[k+4]:=0;
                    if PCXlineBuffer^[i] and 4=4 then
                       ImageBuffer^[k+5]:=15
                    else
                       ImageBuffer^[k+5]:=0;
                    if PCXlineBuffer^[i] and 2=2 then
                       ImageBuffer^[k+6]:=15
                    else
                       ImageBuffer^[k+6]:=0;
                    if PCXlineBuffer^[i] and 1=1 then
                       ImageBuffer^[k+7]:=15
                    else
                       ImageBuffer^[k+7]:=0
                 end
             end { case }
          end;
       PutImage(x,y, ImageBuffer^, BitBlt)
    end;



  procedure SelectShowLineProc(var ShowLine:ShowLineProc);
    begin
       case bits_per_pixel of
          8: ShowLine:=PutImage;
          4: if GetMaxColor=15 then
                ShowLine:=Show16_16
             else
                ShowLine:=Show16_256;
          2: ;  { Not supported! }
          1: case GetMaxColor of
                 1: ShowLine:=PutImage;
                15: ShowLine:=Show2_16;
               255: ShowLine:=Show2_256
             end {case}
       end {case}
    end;



  function  GetVPmaxX : integer;
     var VP : ViewPortType;
     begin
        GetViewSettings(VP);
        GetVPmaxX:=VP.x2-VP.x1
     end;


  function  GetVPmaxY : integer;
     var VP : ViewPortType;
     begin
        GetViewSettings(VP);
        GetVPmaxY:=VP.y2-VP.y1
     end;



  procedure ShowPCX(AutoCentering, PaletteUsage: byte; BitBlt: byte);
    var x1,y1,x2,y2: word;   { Coordinate relative che assumer l'immagine }
        i: word;
        ShowLine: ShowLineProc;
    begin
       if GetGraphMode<0 then RunError(NotInGraphMode);    { La grafica non  stata attivata! }
       if (GetMaxColor=3) or (GetMaxColor>255) then RunError(InvalidGraphMode);
       PCXerrCode:=0;
       if not PCXopen then       { Non  stata usata GetPCXinfo }
          begin
             ReadHeader;
             if PCXresult<>PCXok then exit;
             CheckHeader;
             if PCXresult<>PCXok then exit;
             if header.version in [2,5] then   { C' la palette }
                case header.bits_per_pixel_per_plane*header.Nplanes of
                   4: EvalEGApalette;
                   8: begin
                         ReadColorMap256;
                         if PCXresult<>PCXok then exit
                      end
                end { case }
          end;

       { A questo punto  come se si fosse usata GetPCXinfo }
       bits_per_pixel:=header.Bits_per_pixel_per_plane*header.Nplanes; { Meglio tardi che mai! }
       bytes_per_line:=header.Bytes_per_line_per_plane*header.Nplanes;
       Hres:=header.Xmax-header.Xmin+1;
       Vres:=header.Ymax-header.Ymin+1;

       if (bits_per_pixel=2) or (bits_per_pixel>8) then
          begin
             PCXerrCode:=PCXunknownMemoryModel;
             exit
          end;

       if GetMaxColor+1<(1 shl bits_per_pixel) then   { Colori insufficenti? }
          begin
             PCXerrCode:=PCXnotEnoughColors;
             exit
          end;

       if AutoCentering=active then
          begin
             if Hres-1<GetVPmaxX then     { C' spazio per centrare? }
                begin
                   x1:=(GetVPmaxX-(Hres-1)) div 2;
                   x2:=x1+Hres-1
                end
             else
                begin
                   x1:=0;
                   x2:=GetVPmaxX
                end;
             if Vres-1<GetVPmaxY then     { C' spazio per centrare? }
                begin
                   y1:=(GetVPmaxY-(Vres-1)) div 2;
                   y2:=y1+Vres-1
                end
             else
                begin
                   y1:=0;
                   y2:=GetVPmaxY
                end
          end
       else
          begin                { No AutoCentering.  Giustifica in alto a sx. }
             x1:=0;   y1:=0;
             if Hres-1>GetVPmaxX then   { Si supera il margine destro? }
                x2:=GetVPmaxX
             else
                x2:=x1+Hres-1;
             if Vres-1>GetVPmaxY then   { Si supera il margine inferiore? }
                y2:=GetVPmaxY
             else
                y2:=y1+Vres-1
          end;


       PlaneLength:=(x2-x1+1) div 8;                  {  Verr usato solo con  }
       if (x2-x1+1) mod 8 >0 then inc(PlaneLength);   { immagini a 2 o 16 col. }

       if (bits_per_pixel in [1,4]) and (GetMaxColor=15) then
          PCXlineBufLen:=bytes_per_line + SecurityBlock + PlaneLength*3
       else
          PCXlineBufLen:=bytes_per_line + SecurityBlock;

       if (MaxAvail<PCXlineBufLen) or (PCXlineBufLen>32767) then
          begin
             PCXerrCode:=PCXOutOfMemory;
             exit
          end;
       GetMem(PCXlineBuffer, PCXlineBufLen);

       ImageBufLen:=0;
       if (GetMaxColor=255) and (bits_per_pixel<>8) then    { Serve un buffer di conversione? }
          begin
             ImageBufLen:=(x2-x1+1) + 7 + 4;   { V. nota 5 }
             if MaxAvail<ImageBufLen then
                begin
                   PCXerrCode:=PCXoutOfMemory;
                   FreeMem(PCXlineBuffer, PCXlineBufLen);
                   exit
                end;
             GetMem(ImageBuffer, ImageBufLen)   { Alloca il buffer per la conversione }
          end;


       if (header.version in [2,5]) and (PaletteUsage=active) and (bits_per_pixel<>1) then
          SetPalette(1 shl bits_per_pixel);

       SelectShowLineProc(ShowLine);


       if ImageBufLen=0 then
          begin
             PCXlineBuffer^[-4]:=lo(x2-x1);   PCXlineBuffer^[-3]:=hi(x2-x1);
             PCXlineBuffer^[-2]:=0;   PCXlineBuffer^[-1]:=0
          end
       else
          begin
             ImageBuffer^[-4]:=lo(x2-x1);   ImageBuffer^[-3]:=hi(x2-x1);
             ImageBuffer^[-2]:=0;    ImageBuffer^[-1]:=0
          end;

       index:=0;                                            {    Usati da...    }
       NextByte:=NumBlocksRead;    { No data read in yet }  { ...DecompressLine }
       for i:=y1 to y2 do
          begin
             DecompressLine;
             if PCXerrCode<>0 then   { C' stato un errore di lettura? }
                begin
                   ClosePCX;
                   PCXerrCode:=PCXreadError;  { V. nota 6 }
                   FreeMem(PCXlineBuffer, PCXlineBufLen);
                   if ImageBufLen>0 then FreeMem(ImageBuffer, ImageBufLen);
                   exit
                end;
             ShowLine(x1,i, PCXlineBuffer^, BitBlt)
          end;
       ClosePCX;
       FreeMem(PCXlineBuffer, PCXlineBufLen);
       if ImageBufLen>0 then FreeMem(ImageBuffer, ImageBufLen)
    end;



  procedure GetRGB(DACreg: byte; var R,G,B: byte);
     var reg: registers;
     begin
        reg.ah:=$10;
        reg.al:=$15;
        reg.bx:=DACreg;
        intr($10,reg);
        R:=reg.dh;
        G:=reg.ch;
        B:=reg.cl
     end;


  procedure Get256RGB(var RGB: RGB256Type);
     var reg: registers;
     begin
        reg.ah:=$10;
        reg.al:=$17;
        reg.bx:=$00;
        reg.cx:=256;
        reg.es:=seg(RGB);
        reg.dx:=ofs(RGB);
        intr($10,reg)
     end;



  procedure BGItoPCXplanes;
    var base: word;
    begin
       base:=header.bytes_per_line_per_plane*4;
       move(PCXlineBuffer^[header.bytes_per_line_per_plane*2],
            PCXlineBuffer^[base], header.bytes_per_line_per_plane);
       inc(base, header.bytes_per_line_per_plane);
       move(PCXlineBuffer^[header.bytes_per_line_per_plane],
            PCXlineBuffer^[base], header.bytes_per_line_per_plane);
       inc(base, header.bytes_per_line_per_plane);
       move(PCXlineBuffer^[0],
            PCXlineBuffer^[base], header.bytes_per_line_per_plane)
    end;



  procedure MakePCX(X1, Y1, X2, Y2: word);
     var Palette    : PaletteType;
         gd, gm     : integer;
         R,G,B      : byte;
         c          : ShortInt;
         RLEbuffer  : array[0..BlockBufLen-1] of byte absolute BlockBuffer;
         RLEindex   : word;
         i, j       : word;
         MinI, MaxI : word;
         count      : byte;
         LastByte   : byte;
    begin
       if not PCXassigned then RunError(PCXnotAssigned);
       if GetGraphMode<0 then RunError(NotInGraphMode);
       if (GetMaxColor=3) or (GetMaxColor>255) then RunError(InvalidGraphMode);
       if X2>GetMaxX then X2:=GetMaxX;
       if Y2>GetMaxY then Y2:=GetMaxY;
       if (X1>X2) or (Y1>Y2) then RunError(InvalidCoords);

       PCXerrCode:=0;

       FillChar(header, SizeOf(header), 0);
       header.Manufacturer:=10;
       header.Encoding:=1;
       header.PaletteInfo:=1;
       move(banner, header.Filler, 2);
       move(banner[28], header.Filler[76], 49);
       move(banner, header.Filler[126], 2);

       header.Xmin:=0;
       header.Ymin:=0;
       header.Xmax:=X2-X1;
       header.Ymax:=Y2-Y1;

       case GetMaxColor of
          1: begin
                header.Version:=3;
                header.bits_per_pixel_per_plane:=1;
                header.Nplanes:=1;
                header.Bytes_per_line_per_plane:=(x2-x1+1) div 8;
                if (x2-x1+1) mod 8 <> 0 then inc(header.Bytes_per_line_per_plane);
                header.ColorMap16[1,REDc]:=$FF;    {    Per     }
                header.ColorMap16[1,GREENc]:=$FF;  { soddisfare }
                header.ColorMap16[1,BLUEc]:=$FF    {    QPEG    }
             end;
         15: begin
                header.Version:=5;
                header.bits_per_pixel_per_plane:=1;
                header.Nplanes:=4;
                header.Bytes_per_line_per_plane:=(x2-x1+1) div 8;
                if (x2-x1+1) mod 8 <> 0 then inc(header.Bytes_per_line_per_plane);
                GetPalette(palette);
                DetectGraph(gd,gm);
                if (gd>=3) and (gd<=5) then    { Scheda EGA }
                   for i:=0 to 15 do
                      begin
                         c:=palette.colors[i];
                         R := (c and 4 shl 5) or (c and 32 shl 1);    { V. nota 7 }
                         G := (c and 2 shl 6) or (c and 16 shl 2);
                         B := (c shl 7) or (c and  8 shl 3);
                         with header do
                            begin
                               ColorMap16[i,REDc] := R or (R shr 2) or (R shr 4);
                               ColorMap16[i,GREENc] := G or (G shr 2) or (G shr 4);
                               ColorMap16[i,BLUEc] := B or (B shr 2) or (B shr 4);
                            end
                      end
                else                           { Scheda MCGA/VGA/SVGA }
                   for i:=0 to 15 do
                      begin
                         GetRGB(palette.colors[i], R,G,B);
                         with header do
                            begin
                               ColorMap16[i,REDc]:=R shl 2;
                               ColorMap16[i,GREENc]:=G shl 2;
                               ColorMap16[i,BLUEc]:=B shl 2
                            end
                      end
             end;
        255: begin
                header.Version:=5;
                header.bits_per_pixel_per_plane:=8;
                header.Nplanes:=1;
                header.Bytes_per_line_per_plane:=x2-x1+1;
                Get256RGB(ColorMap256);
                for i:=0 to 255 do
                   begin
                      ColorMap256[i,REDc]:=ColorMap256[i,REDc] shl 2;
                      ColorMap256[i,GREENc]:=ColorMap256[i,GREENc] shl 2;
                      ColorMap256[i,BLUEc]:=ColorMap256[i,BLUEc] shl 2
                   end
             end
       end; { case }


       ClosePCX;            { Se era aperto lo chiude }
       if PCXerrCode<>0 then exit;
       rewrite(PCXfile,1);
       if IOresult<>0 then
          begin
             PCXerrCode:=PCXcannotOpen;
             exit
          end;
       BlockWrite(PCXfile, header, SizeOf(header));
       if IOresult<>0 then
          begin
             Close(PCXfile);
             PCXerrCode:=PCXwriteError;
             exit
          end;


       PCXlineBufLen:=ImageSize(x1,0,x2,0);
       if GetMaxColor=15 then    { Serve altro spazio per la conversione? }
          PCXlineBufLen:= PCXlineBufLen + header.bytes_per_line_per_plane * 3;
       if MaxAvail<PCXlineBufLen then
          begin
             PCXerrCode:=PCXOutOfMemory;
             exit
          end;
       GetMem(PCXlineBuffer, PCXlineBufLen);


       if GetMaxColor=15 then
          MinI:=header.bytes_per_line_per_plane*3+1
       else MinI:=1;
       MaxI:=PCXlineBufLen-4-1;
       RLEindex:=0;

       for j:=Y1 to Y2 do
          begin
             GetImage(X1,j, X2,j, PCXlineBuffer^);
             if GetMaxColor=15 then BGItoPCXplanes;

             LastByte:=PCXlineBuffer^[MinI-1];
             i:=MinI;        { Indice in PCXlineBuffer^[] }
             count:=1;       { Contatore ripetizioni }

             repeat
                if (PCXlineBuffer^[i]=LastByte) and (count<63) and (i<=MaxI) then
                   begin               { C' una ripetizione }
                      inc(count);   inc(i)
                   end
                else                   { Fine ripetizioni }
                   begin
                      if count>1 then
                         begin
                            RLEbuffer[RLEindex]:=count or 192;
                            inc(RLEindex);
                            RLEbuffer[RLEindex]:=LastByte;
                            inc(RLEindex)
                         end
                      else
                         begin
                            if LastByte and 192 = 192 then   { Ripetizione singola }
                               begin
                                  RLEbuffer[RLEindex]:=count or 192;
                                  inc(RLEindex);
                                  RLEbuffer[RLEindex]:=LastByte;
                                  inc(RLEindex)
                               end
                            else                             { Non c' ripetizione }
                               begin
                                  RLEbuffer[RLEindex]:=LastByte;
                                  inc(RLEindex)
                               end
                         end;
                      LastByte:=PCXlineBuffer^[i];
                      inc(i);
                      count:=1
                   end
             until i=MaxI+2;

             if RLEindex >= 512 then
                begin
                   BlockWrite(PCXfile, RLEbuffer, RLEindex);
                   if IOresult<>0 then
                      begin
                         ClosePCX;
                         PCXerrCode:=PCXwriteError;
                         FreeMem(PCXlineBuffer, PCXlineBufLen);
                         exit
                      end
                   else  RLEindex:=0
                end
          end;

       if RLEindex <> 0 then
          begin
             BlockWrite(PCXfile, RLEbuffer, RLEindex);
             if IOresult<>0 then
                begin
                   ClosePCX;
                   PCXerrCode:=PCXwriteError;
                   FreeMem(PCXlineBuffer, PCXlineBufLen);
                   exit
                end
          end;

       if GetMaxColor=255 then
          begin
             i:=12;   { 12 = Palette indicator }
             BlockWrite(PCXfile, i, 1);                { Write palette indicator }
             BlockWrite(PCXfile, ColorMap256, 256*3);  { Write 256-colors palette }
             if IOresult<>0 then
                begin
                   ClosePCX;
                   PCXerrCode:=PCXwriteError;
                   FreeMem(PCXlineBuffer, PCXlineBufLen);
                   exit
                end
          end;

       ClosePCX;
       FreeMem(PCXlineBuffer, PCXlineBufLen)
    end;





begin  { Initialization }
   PCXassigned:=false;
   PCXopen:=false;
   PCXerrCode:=0;
   banner[1]:=#13   { Per assicurare il linkaggio del banner. }
end.


(* ---------------------------- Note sul sorgente ---------------------------

1) Manca l'overflow checking e manca il test per verificare che x2>=x1 e
   y2>=y1, ma poco importa perch la funzione  per uso interno e non c'
   pericolo.
2) La procedura Close accede al disco solo se il file  stato aperto in
   scrittura, perci solo in questo caso si pu verificare un errore di I/O.
   Solo MakePCX opera in scrittura, quindi questo errore pu verficarsi solo
   quando ClosePCX viene chiamata internamente da MakePCX; in tutti gli altri
   casi ClosePCX genera il codice PCXok.
3) Tecnica di conversione RGB -> EGA:
   si prendono i due bit pi significativi dei registri RGB (li indichiamo
   con R-r, G-g e B-b, dove le maiuscole indicano i bit 7 e le minuscole i
   bit 6) e si costruisce la stringa "00rgbRGB".
4) I 4 piani di bit si trovano inizialmente nell'ordine 1-2-3-4 e vengono
   portati nell'ordine 4-3-2-1.  La variabile "base" punta al primo byte
   libero dopo il piano 4.  Vengono copiati solo i piani 1, 2 e 3.
5) Il 7 serve perch l'algoritmo di conversione converte pi pixel di quanti
   ce ne sono in realt (se l'ultimo byte della bitmap decompressa non 
   "pieno").  Il 4  per altezza e largh. dell'immagine.
6) Ripristina il codice di errore che pu essere azzerato da ClosePCX.
7) Tecnica di conversione EGA -> RGB:
   dalla palette EGA si ricavano i due bit pi significativi (il 7 e il 6)
   dei registri RGB e poi si copiano questi due bit nelle due coppie seguenti
   (5-4 e 3-2).  I due bit meno significativi vengono lasciati a zero.
   Si ottengono stringhe del tipo "RrRrRr00", "GgGgGg00" e "BbBbBb00", come
   quelle impostate per default dal BIOS.

-------------------------------------------------------------------------- *)
